@@ -23,6 +23,7 @@
  */
 package hudson.remoting;
 
+import edu.umd.cs.findbugs.annotations.SuppressWarnings;
 import hudson.remoting.CommandTransport.CommandReceiver;
 import hudson.remoting.ExportTable.ExportList;
 import hudson.remoting.PipeWindow.Key;
@@ -31,6 +32,7 @@ import hudson.remoting.forward.ListeningPort;
 import hudson.remoting.forward.ForwarderFactory;
 import hudson.remoting.forward.PortForwarder;
 
+import java.io.Closeable;
 import java.io.IOException;
 import java.io.InputStream;
 import java.io.ObjectOutputStream;
@@ -105,7 +107,7 @@ import java.util.concurrent.ThreadFactory;
  *
  * @author Kohsuke Kawaguchi
  */
-public class Channel implements VirtualChannel, IChannel {
+public class Channel implements VirtualChannel, IChannel, Closeable {
     private final CommandTransport transport;
 
     /**
@@ -593,6 +595,13 @@ public class Channel implements VirtualChannel, IChannel {
      * Classloading will consult the preloaded jars before performing
      * network transfer of class files.
      *
+     * <p><strong>Beware</strong> that this method is not useful in all configurations.
+     * If a {@link RemoteClassLoader} has another {@link RemoteClassLoader} as a
+     * {@linkplain ClassLoader#getParent parent}, which would be typical, then preloading
+     * a JAR in it will not reduce network round-trips: each class load still has to call
+     * {@link ClassLoader#loadClass(String, boolean) loadClass} on the parent, which will
+     * wind up checking the remote side just to get a negative answer.
+     *
      * @param classLoaderRef
      *      This parameter is used to identify the remote classloader
      *      that will prefetch the specified jar files. That is, prefetching
@@ -705,33 +714,39 @@ public class Channel implements VirtualChannel, IChannel {
      * @param e
      *      The error that caused the connection to be aborted. Never null.
      */
-    protected synchronized void terminate(IOException e) {
-        if (e==null)    throw new IllegalArgumentException();
-        outClosed=inClosed=e;
+    @java.lang.SuppressWarnings("ToArrayCallWithZeroLengthArrayArgument")
+    @SuppressWarnings("ITA_INEFFICIENT_TO_ARRAY") // intentionally; race condition on listeners otherwise
+    protected void terminate(IOException e) {
         try {
-            transport.closeRead();
-        } catch (IOException x) {
-            logger.log(Level.WARNING, "Failed to close down the reader side of the transport",x);
-        }
-        try {
-            synchronized(pendingCalls) {
-                for (Request<?,?> req : pendingCalls.values())
-                    req.abort(e);
-                pendingCalls.clear();
-            }
-            synchronized (executingCalls) {
-                for (Request<?, ?> r : executingCalls.values()) {
-                    java.util.concurrent.Future<?> f = r.future;
-                    if(f!=null) f.cancel(true);
+            synchronized (this) {
+                if (e == null) throw new IllegalArgumentException();
+                outClosed = inClosed = e;
+                try {
+                    transport.closeRead();
+                } catch (IOException x) {
+                    logger.log(Level.WARNING, "Failed to close down the reader side of the transport", x);
                 }
-                executingCalls.clear();
-            }
+                try {
+                    synchronized (pendingCalls) {
+                        for (Request<?, ?> req : pendingCalls.values())
+                            req.abort(e);
+                        pendingCalls.clear();
+                    }
+                    synchronized (executingCalls) {
+                        for (Request<?, ?> r : executingCalls.values()) {
+                            java.util.concurrent.Future<?> f = r.future;
+                            if (f != null) f.cancel(true);
+                        }
+                        executingCalls.clear();
+                    }
+                } finally {
+                    notifyAll();
+                }
+            } // JENKINS-14909: leave synch block
         } finally {
-            notifyAll();
-
-            if (e instanceof OrderlyShutdown)   e = null;
-            for (Listener l : listeners.toArray(new Listener[listeners.size()]))
-                l.onClosed(this,e);
+            if (e instanceof OrderlyShutdown) e = null;
+            for (Listener l : listeners.toArray(new Listener[0]))
+                l.onClosed(this, e);
         }
     }
 
@@ -825,8 +840,8 @@ public class Channel implements VirtualChannel, IChannel {
      * {@link ObjectOutputStream}, and it's the last command to be read.
      */
     private static final class CloseCommand extends Command {
-        private CloseCommand(Throwable cause) {
-            super(cause);
+        private CloseCommand(Channel channel, Throwable cause) {
+            super(channel,cause);
         }
 
         protected void execute(Channel channel) {
@@ -891,7 +906,7 @@ public class Channel implements VirtualChannel, IChannel {
     public synchronized void close(Throwable diagnosis) throws IOException {
         if(outClosed!=null)  return;  // already closed
 
-        send(new CloseCommand(diagnosis));
+        send(new CloseCommand(this,diagnosis));
         outClosed = new IOException().initCause(diagnosis);   // last command sent. no further command allowed. lock guarantees that no command will slip inbetween
         try {
             transport.closeWrite();
@@ -1017,8 +1032,10 @@ public class Channel implements VirtualChannel, IChannel {
      * @return
      */
     public ListeningPort createLocalToRemotePortForwarding(int recvPort, String forwardHost, int forwardPort) throws IOException, InterruptedException {
-        return new PortForwarder( recvPort,
-            ForwarderFactory.create(this,forwardHost,forwardPort));
+        PortForwarder portForwarder = new PortForwarder(recvPort,
+                ForwarderFactory.create(this, forwardHost, forwardPort));
+        portForwarder.start();
+        return portForwarder;
     }
 
     /**
