@@ -31,6 +31,7 @@ import org.kohsuke.args4j.Option;
 import org.kohsuke.args4j.CmdLineParser;
 import org.kohsuke.args4j.CmdLineException;
 
+import javax.crypto.spec.IvParameterSpec;
 import javax.xml.parsers.DocumentBuilder;
 import javax.xml.parsers.DocumentBuilderFactory;
 import javax.xml.parsers.ParserConfigurationException;
@@ -41,13 +42,17 @@ import javax.net.ssl.SSLContext;
 import javax.net.ssl.X509TrustManager;
 import javax.net.ssl.HostnameVerifier;
 import javax.net.ssl.SSLSession;
+import java.io.FileOutputStream;
 import java.io.IOException;
 import java.io.InputStream;
 import java.io.OutputStream;
 import java.io.File;
 import java.io.BufferedInputStream;
 import java.io.BufferedOutputStream;
+import java.io.ByteArrayInputStream;
+import java.io.ByteArrayOutputStream;
 import java.io.FileWriter;
+import java.io.PrintStream;
 import java.lang.reflect.InvocationTargetException;
 import java.lang.reflect.Method;
 import java.net.URL;
@@ -59,6 +64,7 @@ import java.net.InetSocketAddress;
 import java.net.HttpURLConnection;
 import java.net.Authenticator;
 import java.net.PasswordAuthentication;
+import java.security.GeneralSecurityException;
 import java.util.ArrayList;
 import java.util.List;
 import java.util.Locale;
@@ -70,8 +76,8 @@ import java.security.NoSuchAlgorithmException;
 import java.security.KeyManagementException;
 import java.security.SecureRandom;
 import java.util.Properties;
-
-import org.apache.commons.codec.binary.Base64;
+import javax.crypto.Cipher;
+import javax.crypto.spec.SecretKeySpec;
 
 /**
  * Entry point for running a {@link Channel}. This is the main method of the slave JVM.
@@ -89,6 +95,9 @@ public class Launcher {
     @Option(name="-ping")
     public boolean ping = true;
 
+    @Option(name="-slaveLog", usage="create local slave error log")
+    public File slaveLog = null;
+
     @Option(name="-text",usage="encode communication with the master with base64. " +
             "Useful for running slave over 8-bit unsafe protocol like telnet")
     public void setTextMode(boolean b) {
@@ -104,6 +113,9 @@ public class Launcher {
     @Option(name="-jnlpCredentials",metaVar="USER:PASSWORD",usage="HTTP BASIC AUTH header to pass in for making HTTP requests.")
     public String slaveJnlpCredentials = null;
 
+    @Option(name="-secret", metaVar="HEX_SECRET", usage="Slave connection secret to use instead of -jnlpCredentials.")
+    public String secret;
+
     @Option(name="-cp",aliases="-classpath",metaVar="PATH",
             usage="add the given classpath elements to the system classloader.")
     public void addClasspath(String pathList) throws Exception {
@@ -175,6 +187,9 @@ public class Launcher {
     }
 
     public void run() throws Exception {
+        if (slaveLog!=null) {
+            System.setErr(new PrintStream(new TeeOutputStream(System.err,new FileOutputStream(slaveLog))));
+        }
         if(auth!=null) {
             final int idx = auth.indexOf(':');
             if(idx<0)   throw new CmdLineException(null, "No ':' in the -auth option");
@@ -214,11 +229,17 @@ public class Launcher {
     public List<String> parseJnlpArguments() throws ParserConfigurationException, SAXException, IOException, InterruptedException {
         while (true) {
             try {
+                if (secret != null) {
+                    slaveJnlpURL = new URL(slaveJnlpURL + "?encrypt=true");
+                    if (slaveJnlpCredentials != null) {
+                        throw new IOException("-jnlpCredentials and -secret are mutually exclusive");
+                    }
+                }
                 URLConnection con = slaveJnlpURL.openConnection();
                 if (con instanceof HttpURLConnection && slaveJnlpCredentials != null) {
                     HttpURLConnection http = (HttpURLConnection) con;
                     String userPassword = slaveJnlpCredentials;
-                    String encoding = new String(new Base64().encodeBase64(userPassword.getBytes()));
+                    String encoding = Base64.encode(userPassword.getBytes());
                     http.setRequestProperty("Authorization", "Basic " + encoding);
                 }
                 con.connect();
@@ -234,17 +255,34 @@ public class Launcher {
 
                 // check if this URL points to a .jnlp file
                 String contentType = con.getHeaderField("Content-Type");
-                if(contentType==null || !contentType.startsWith("application/x-java-jnlp-file")) {
+                String expectedContentType = secret == null ? "application/x-java-jnlp-file" : "application/octet-stream";
+                InputStream input = con.getInputStream();
+                if (secret != null) {
+                    byte[] payload = toByteArray(input);
+                    // the first 16 bytes (128bit) are initialization vector
+
+                    try {
+                        Cipher cipher = Cipher.getInstance("AES/CFB8/NoPadding");
+                        cipher.init(Cipher.DECRYPT_MODE,
+                                new SecretKeySpec(fromHexString(secret.substring(0, Math.min(secret.length(), 32))), "AES"),
+                                new IvParameterSpec(payload,0,16));
+                        byte[] decrypted = cipher.doFinal(payload,16,payload.length-16);
+                        input = new ByteArrayInputStream(decrypted);
+                    } catch (GeneralSecurityException x) {
+                        throw (IOException)new IOException("Failed to decrypt the JNLP file. Invalid secret key?").initCause(x);
+                    }
+                }
+                if(contentType==null || !contentType.startsWith(expectedContentType)) {
                     // load DOM anyway, but if it fails to parse, that's probably because this is not an XML file to begin with.
                     try {
-                        dom = loadDom(slaveJnlpURL, con);
+                        dom = loadDom(slaveJnlpURL, input);
                     } catch (SAXException e) {
                         throw new IOException(slaveJnlpURL+" doesn't look like a JNLP file; content type was "+contentType);
                     } catch (IOException e) {
                         throw new IOException(slaveJnlpURL+" doesn't look like a JNLP file; content type was "+contentType);
                     }
                 } else {
-                    dom = loadDom(slaveJnlpURL, con);
+                    dom = loadDom(slaveJnlpURL, input);
                 }
 
                 // exec into the JNLP launcher, to fetch the connection parameter through JNLP.
@@ -277,9 +315,26 @@ public class Launcher {
         }
     }
 
-    private static Document loadDom(URL slaveJnlpURL, URLConnection con) throws ParserConfigurationException, SAXException, IOException {
+    private byte[] toByteArray(InputStream input) throws IOException {
+        ByteArrayOutputStream baos = new ByteArrayOutputStream();
+        int c;
+        while ((c = input.read()) != -1) {
+            baos.write(c);
+        }
+        return baos.toByteArray();
+    }
+
+    // from hudson.Util
+    private static byte[] fromHexString(String data) {
+        byte[] r = new byte[data.length() / 2];
+        for (int i = 0; i < data.length(); i += 2)
+            r[i / 2] = (byte) Integer.parseInt(data.substring(i, i + 2), 16);
+        return r;
+    }
+
+    private static Document loadDom(URL slaveJnlpURL, InputStream is) throws ParserConfigurationException, SAXException, IOException {
         DocumentBuilder db = DocumentBuilderFactory.newInstance().newDocumentBuilder();
-        return db.parse(con.getInputStream(),slaveJnlpURL.toExternalForm());
+        return db.parse(is, slaveJnlpURL.toExternalForm());
     }
 
     /**
@@ -311,6 +366,12 @@ public class Launcher {
     }
 
     private void runOnSocket(Socket s) throws IOException, InterruptedException {
+        // this prevents a connection from silently terminated by the router in between or the other peer
+        // and that goes without unnoticed. However, the time out is often very long (for example 2 hours
+        // by default in Linux) that this alone is enough to prevent that.
+        s.setKeepAlive(true);
+        // we take care of buffering on our own
+        s.setTcpNoDelay(true);
         main(new BufferedInputStream(new SocketInputStream(s)),
              new BufferedOutputStream(new SocketOutputStream(s)), mode,ping);
     }
